import rapidjson
from xaal.lib import bindings

"""
Since version 0.7 xAAL msg are CBOR encoded, this means w/ addresses now
use Cbor UUID tag. So to avoid error w/ JSON encoder of UUID, we need to 
use rapidjson that provide UUID to JSON. This wrapper module provide this
to sio package.
"""

def dumps(*args,**kwargs):
    obj = args[0]
    try:
        r = rapidjson.dumps(obj,uuid_mode=rapidjson.UM_CANONICAL)        
    except TypeError:
        import pdb;pdb.set_trace()
    return r


def cleanup(obj):
    """ 
    recursive walk a object to search for un-wanted CBOR tags.
    Transform this tag in string format, this can be UUID, URL..

    /!\ use with caution, this operate in-place changes.

    This is quite the same code as cbor.cleanup() but simply call
    get() instead of str() 
    """
    if isinstance(obj,list):
        for i in range(0,len(obj)):
            obj[i] = cleanup(obj[i]) 
        return obj

    if isinstance(obj,dict):
        for k in obj.keys():
            obj.update({k:cleanup(obj[k])})
        return obj

    if type(obj) in bindings.classes:
        r = obj.get()
        return r
    else:
        return obj

def prepare_cbor(obj):
    """ transfor a cbor object into something that rapdjson can handle """
    import copy
    r = copy.deepcopy(obj)
    return cleanup(r)

loads = rapidjson.loads




from xaal.lib import tools
from .default import view,route
from .default import xaal_core as xaal


from bottle import get,response
import logging
import json
import requests

TOKEN=None
URL=None
CLASS = None


DAILY_REQ = """ [ $TOKEN '%s' { 'devid' '%s'  } NOW 49 h ] FETCH """

bindings = {'thermometer.basic' : 'temperature',
            'barometer.basic'   : 'pressure',
            'hygrometer.basic'  : 'humidity',
            'powermeter.basic'  : 'power',
            'co2meter.basic'  : 'co2',
            'luxmeter.basic'  : 'illuminance',
            }

logger = logging.getLogger(__name__)

def setup():
    global TOKEN,URL,CLASS
    cfg = tools.load_cfg(xaal.PACKAGE_NAME).get('config',None)
    if cfg:
        TOKEN = cfg.get('warp10_token',None)
        URL = cfg.get('warp10_url',None)
        CLASS = cfg.get('warp10_class',None)
        logger.warning("%s %s" % (URL,CLASS))

def init_req():
    return """'%s' 'TOKEN' STORE\n""" % (TOKEN)


def query(req):
    r = requests.post(URL,req)
    print(r)
    return r.text
    

def to_chartjs(data):
    r = []
    for tuple_ in data:
        date = round(tuple_[0] / 1000)
        value = tuple_[1]
        #r.append({"x": date, "y":value })
        r.append((date,value))
    return r

@get('/warp10/daily/<addr>')
def daily(addr):
    global CLASS
    response.headers['Content-Type'] = 'application/json'
    res = {}
    #import pdb;pdb.set_trace()
    dev=xaal.monitor.devices.get_with_addr(tools.get_uuid(addr))
    if dev:
        req = init_req()
        devtype = dev.dev_type
        var = bindings.get(devtype,None)
        if not var:
            return '{}'
        class_ = '%s.%s.%s' % (CLASS,devtype,var)
        tmp = DAILY_REQ % (class_,addr)
        req = init_req() + tmp
        data=query(req)
        
        res = json.loads(data)[0][0]['v']
    return json.dumps(to_chartjs(res))


@route('/warp10/graph/<addr>')
@view('graph')
def graph(addr):
    r = {"title" : "Warp10 daily","addr":addr}
    return r



from xaal.lib import tools,Engine,Device,helpers
from xaal.monitor import Monitor
from . import sio

import platform
import gevent

import time
import logging

PACKAGE_NAME = "xaal.dashboard"
logger = logging.getLogger(PACKAGE_NAME)

# we use this global variable to share data with greenlet
monitor = None
config = {}

# used for uptime
started = time.time()

def monitor_filter(msg):
    """ Filter incomming messages. Return False if you don't want to use this device"""
    if msg.dev_type.startswith('cli.'):
        return False
    return True

def event_handler(ev_type,dev):
    logger.debug("Event %s %s" % (ev_type,dev.address))
    msg = { 'address': dev.address.get(),'attributes':dev.attributes}
    # FIXME: This should be in core.sio
    sio.broadcast('event_attributeChanges',msg)
    
def setup(engine):
    """ setup xAAL Engine & Device. And start it in a Greenlet"""
    global monitor,config
    helpers.setup_console_logger()
    cfg = tools.load_cfg(PACKAGE_NAME)
    if not cfg:
        logger.info('Missing config file, building a new one')
        cfg = tools.new_cfg(PACKAGE_NAME)
        cfg.write()
    dev            = Device("hmi.dashboard")
    dev.address    = tools.get_uuid(cfg['config']['addr'])
    dev.vendor_id  = "IHSEV"
    dev.product_id = "WEB Interface"
    dev.version    = 0.1
    dev.info       = "%s@%s" % (PACKAGE_NAME,platform.node())

    engine.add_device(dev)
    db_server = None
    if 'db_server' in cfg['config']:
        db_server = cfg['config']['db_server']
    else:
        logger.info('You can set "db_server" in the config file')
    monitor = Monitor(dev,filter_func=monitor_filter,db_server=tools.get_uuid(db_server))
    monitor.subscribe(event_handler)
    config = cfg

def xaal_loop(engine):
    """ xAAL Engine Loop Greenlet"""
    engine.start()
    while 1:
        engine.loop()

def send_request(addr,action,body):
    eng = monitor.dev.engine
    eng.send_request(monitor.dev,[addr],action,body)       

def get_uptime():
    return int(time.time() - started)

def get_device(addr):
    uuid = tools.get_uuid(addr)
    return monitor.devices.get_with_addr(uuid)

def update_kv(addr,kv):
    dev = get_device(addr)
    db_server = monitor.db_server
    if dev:
        body = {'device':dev.address,'map':kv}
        print(body)
        send_request(db_server,'update_keys_values',body)
        dev.set_db(kv)


import socketio
from . import xaal_core
from . import json
from xaal.lib import tools

socket = socketio.Server(async_mode='gevent',ping_interval=0.5,json=json,cors_allowed_origins='*',logger=False)

def setup(wsgi_app):
    global socket
    app = socketio.Middleware(socket, wsgi_app)
    return app

def remote_addr(sid):
    env = socket.environ[sid]
    return env['REMOTE_ADDR']

def broadcast(event,sio_msg):
    socket.emit(event,json.prepare_cbor(sio_msg))

@socket.on('send_request')
def send_xaal_request(sid,addr,action,body):
    xaal_core.send_request(tools.get_uuid(addr),action,body)

@socket.on('query_attributes')
def query_attributes(sid,addr):
    dev = xaal_core.get_device(addr)
    if dev:
        msg = { 'address': dev.address.get(),'attributes':json.prepare_cbor(dev.attributes)}
        socket.emit('event_attributeChanges',msg,room=sid)

@socket.on('refresh_attributes')
def refresh_attributes(sid,addrs):
    print("[%s] Refresh attributes" % remote_addr(sid))
    for addr in addrs:
        dev = xaal_core.get_device(addr)
        if dev:
            msg = { 'address': dev.address.get(),'attributes':json.prepare_cbor(dev.attributes)}
            socket.emit('event_attributeChanges',msg,room=sid)
        else:
            print("Unknow device: %s" % addr)

@socket.on('debug')
def debug(sid):
    import pdb;pdb.set_trace()

@socket.on('ping_from_client')
def ping(sid):
    """ used for the latency test only"""
    socket.emit('pong_from_server', room=sid)



from email.policy import default
from gevent import monkey; monkey.patch_all(thread=False)
from gevent.pywsgi import WSGIServer
from geventwebsocket.handler import WebSocketHandler
import gevent

from bottle import Bottle,debug,get,redirect,static_file,TEMPLATE_PATH,default_app

# load pages 
from . import pages
# load components
from .pages import warp10
from .core import xaal_core
from .core import sio 

import os

HOME = os.path.dirname(__file__)
PORT = 9090

server = None

@get('/static/<filename:path>')
def send_static(filename):
    root = os.path.join(HOME,'static')
    return static_file(filename, root=root)

@get('/')
def goto_home():
    redirect('/grid')
    
def run():
    """ start the xAAL stack & launch the HTTP stuff"""
    global server
    # add the default template directory to the bottle search path
    root = os.path.join(HOME,'templates')
    TEMPLATE_PATH.append(root)
    warp10.setup()
    # debug disable template cache & enable error reporting
    debug(True)
    #bottle_app = Bottle()
    #app = sio.setup(bottle_app)
    app = sio.setup(default_app())
    #server = WSGIServer(("", 9090), app, handler_class=WebSocketHandler, keyfile=KEY_PATH+'key.pem', certfile=KEY_PATH+'cert.pem')
    server = WSGIServer(f":{PORT}", app, handler_class=WebSocketHandler)
    server.serve_forever()

def stop():
    global server
    server.stop()

def main():
    # This apps can run with gevent alone (without asyncio)   
    # bottle.py use gevent to handle websocket. 
    # I maintain this to be able to run it on older 2.7 Python
    from xaal.lib import Engine
    engine = Engine()
    xaal_core.setup(engine)
    green_let = gevent.Greenlet(xaal_core.xaal_loop, engine)
    green_let.start()
    try:
        run()
    except KeyboardInterrupt:
        stop()
        print("Bye Bye...")

def setup(engine):
    # This is a AsyncEngine only setup, used w/ xaal-pkgrun for example
    xaal_core.setup(engine)
    gevent.spawn(run)
    engine.on_stop(stop)
    return True

if __name__ == '__main__':
    main()



from bottle import route,request
from ..core import xaal_core
from bottle import mako_template as template
import functools


def build_cookie_token(profile,key):
    import hashlib
    ua = request.environ.get('HTTP_USER_AGENT')
    buf = '%s/%s/%s' % (ua,profile,key)
    h = hashlib.blake2b(buf.encode('utf-8')).hexdigest()
    return h

def search_token_profile(token):
    """search the profile for a given token"""
    profiles = xaal_core.config.get('profiles',{})
    for p in profiles:
        for k in profiles[p].get('keys',[]):
            h = build_cookie_token(p,k)
            if h == token:
                return profiles[p]
    return {}

def get_config_profile(name):
    return xaal_core.config.get('profiles',{}).get(name,{})

def get_profile():
    return search_token_profile( request.get_cookie('token') )

def view(tpl_name,**defaults):
    # this is a cut / paste from the bottle view. 
    # just tweak it to support custom template from profile
    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, dict):
                tplvars = defaults.copy()
                tplvars.update(result)
                profile = get_profile()
                tmpl = profile.get('%s_tmpl' %tpl_name,'%s.mako' % tpl_name)
                tplvars.update({'profile':profile})
                return template(tmpl, **tplvars)
            elif result is None:
                return template(tpl_name, defaults)
            return result

        return wrapper

    return decorator



from .default import view,route,xaal_core

from bottle import request,response,redirect,get,post

import copy

@route('/stats')
@view('stats')
def stats():
    total = 0
    results = {}
    for dev in xaal_core.monitor.devices:
        total = total + 1
        try:
            k = dev.dev_type
            results[k]=results[k]+1
        except KeyError:
            results.update({k:1})
    r = {"title" : "Network stats"}
    r.update({"total"   :total})
    r.update({"dev_types":results})
    r.update({"uptime"  : xaal_core.get_uptime()})
    return r


@route('/bottle')
@view('bottle')
def info():
    r = {"title" : "Bottle Server Info"}
    r.update({"headers" : request.headers})
    r.update({"query"   : request.query})
    r.update({"environ" : copy.copy(request.environ)})
    return r



@route('/devices')
@view('devices')
def get_devices():
    r = {"title" : "devices list","active_menu":"devices"}
    devs = xaal_core.monitor.devices
    r.update({"devs" : devs})
    return r


@route('/generic/<addr>')
@view('generic')
def get_device(addr):
    r = {"title" : "device %s" % addr}
    dev = xaal_core.get_device(addr)
    if dev:
        r.update({"dev" : dev})
    return r

@get('/edit_metadata/<addr>')
@view('edit_metadata')
def edit_metadata(addr):
    r = {"title" : "device %s" % addr}
    dev = xaal_core.get_device(addr)
    if dev:
        r.update({"dev" : dev})
    return r


@post('/edit_metadata/<addr>')
@view('edit_metadata')
def save_metadata(addr):
    form = dict(request.forms.decode()) # just to shut up lint
    kv = {}
    for k in form:
        key = str(k)
        if form[k]=='': continue
        if key.startswith('key_'):
            id_ = key.split('key_')[-1]
            v_key = 'value_'+id_
            if v_key in form:
                if form[v_key] =='':
                    value = None
                else:
                    value = form[v_key]
                kv.update({form[k]:value})
    xaal_core.update_kv(addr,kv)
    return edit_metadata(addr)
    



@route('/grid')
@view('grid')
def test_grid():
    from xaal.lib import tools
    return {"title" : "Grid","devices":xaal_core.monitor.devices,"tools":tools,"active_menu":"grid"}


@route('/latency')
def socketio_latency_test():
    redirect('/static/latency.html')


@route('/links')
@view('links')
def links():
    return {'title':'Links'}


@get('/token/<profile>/<key>')
def set_token(profile,key):
    from .default import get_config_profile,build_cookie_token
    import time

    profile_cfg = get_config_profile(profile)
    if profile_cfg != {}:
        if key in profile_cfg.get('keys',[]):
            tmp =  build_cookie_token(profile,key)
            expires = time.time() + 60*60*24*365*2
            response.set_cookie('token',tmp,path='/',expires=expires)
        else:
            print("Invalid key")
    redirect('/')
